@@ -1,4 +1,5 @@
 {-# LANGUAGE BangPatterns               #-}
+{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE NoMonomorphismRestriction  #-}
@@ -33,6 +34,11 @@ import           Text.Printf
 import qualified Text.XmlHtml                       as X
 import qualified Text.XmlHtml.HTML.Meta             as X
 ------------------------------------------------------------------------------
+#if !MIN_VERSION_base(4,8,0)
+import           Data.Foldable                      (Foldable)
+#endif
+import qualified Data.Foldable                      as Foldable
+------------------------------------------------------------------------------
 import           Heist.Common
 import           Heist.Internal.Types.HeistState
 ------------------------------------------------------------------------------
@@ -45,7 +51,7 @@ import           Heist.Internal.Types.HeistState
 -- The more interesting part of the type signature is what comes before the
 -- return value.  The first type parameter in @'HeistT' n IO@ is the runtime
 -- monad.  This reveals that the Chunks know about the runtime monad.  The
--- second type parameter in @HeistT n IO@ is @IO@.  This tells is that the
+-- second type parameter in @HeistT n IO@ is @IO@.  This tells us that the
 -- compiled splices themselves are run in the IO monad, which will usually
 -- mean at load time.  Compiled splices run at load time, and they return
 -- computations that run at runtime.
@@ -159,17 +165,26 @@ compileTemplate tpath df = do
 ------------------------------------------------------------------------------
 compileTemplates
     :: Monad n
-    => HeistState n
+    => (TPath -> Bool)
+    -> HeistState n
     -> IO (Either [String] (HeistState n))
-compileTemplates hs = do
-    (tmap, hs') <- runHeistT compileTemplates' (X.TextNode "") hs
+compileTemplates f hs = do
+    (tmap, hs') <- runHeistT (compileTemplates' f) (X.TextNode "") hs
     let pre = _splicePrefix hs'
+    let canError = _errorNotBound hs'
+    let errs = _spliceErrors hs'
     let nsErr = if not (T.null pre) && (_numNamespacedTags hs' == 0)
                   then Left [noNamespaceSplicesMsg $ T.unpack pre]
                   else Right ()
-    return $ case _spliceErrors hs' of
-               [] -> nsErr >> (Right $! hs { _compiledTemplateMap = tmap })
-               es -> Left $ either (++) (const id) nsErr $ map T.unpack es
+    return $ if canError
+               then case errs of
+                     [] -> nsErr >>
+                           (Right $! hs { _compiledTemplateMap = tmap })
+                     es -> Left $ either (++) (const id) nsErr $
+                           map (T.unpack . spliceErrorText) es
+               else nsErr >> (Right $! hs { _compiledTemplateMap = tmap
+                                          , _spliceErrors = errs
+                                          })
 
 
 ------------------------------------------------------------------------------
@@ -187,11 +202,12 @@ noNamespaceSplicesMsg pre = unwords
 ------------------------------------------------------------------------------
 compileTemplates'
     :: Monad n
-    => HeistT n IO (H.HashMap TPath ([Chunk n], MIMEType))
-compileTemplates' = do
+    => (TPath -> Bool)
+    -> HeistT n IO (H.HashMap TPath ([Chunk n], MIMEType))
+compileTemplates' f = do
     hs <- getHS
     let tpathDocfiles :: [(TPath, DocumentFile)]
-        tpathDocfiles = map (\(a,b) -> (a, b))
+        tpathDocfiles = filter (f . fst)
                             (H.toList $ _templateMap hs)
     foldM runOne H.empty tpathDocfiles
   where
@@ -257,11 +273,9 @@ lookupSplice :: Text -> HeistT n IO (Maybe (Splice n))
 lookupSplice nm = do
     pre <- getsHS _splicePrefix
     res <- getsHS (H.lookup nm . _compiledSpliceMap)
-    canError <- getsHS _errorNotBound
     if isNothing res && T.isPrefixOf pre nm && not (T.null pre)
       then do
-          when canError $
-            tellSpliceError $ "No splice bound for " `mappend` nm
+          tellSpliceError $ "No splice bound for " `mappend` nm
           return Nothing
       else return res
 
@@ -272,14 +286,18 @@ lookupSplice nm = do
 -- compileNode to generate the appropriate runtime computation.
 runNode :: Monad n => X.Node -> Splice n
 runNode node = localParamNode (const node) $ do
-    pre <- getsHS _splicePrefix
+    hs <- getHS
+    let pre = _splicePrefix hs
     let hasPrefix = (T.isPrefixOf pre `fmap` X.tagName node) == Just True
     when (not (T.null pre) && hasPrefix) incNamespacedTags
     isStatic <- subtreeIsStatic node
     markup <- getsHS _curMarkup
     if isStatic
       then return $! yieldPure $! renderFragment markup [parseAttrs node]
-      else compileNode node
+      else localHS (\hs' -> hs' {_splicePath =
+                                 (_curContext hs', _curTemplateFile hs',
+                                  X.elementTag node):(_splicePath hs')}) $
+           compileNode node
 
 
 parseAttrs :: X.Node -> X.Node
@@ -613,9 +631,11 @@ textSplice f = fromText . f
 
 
 ------------------------------------------------------------------------------
--- | Converts a pure Node splice function to a pure Builder splice function.
+-- | This is the same as htmlNodeSplice.  
 nodeSplice :: (a -> [X.Node]) -> a -> Builder
 nodeSplice f = X.renderHtmlFragment X.UTF8 . f
+{-# DEPRECATED nodeSplice
+   "Use xmlNodeSplice or htmlNodeSplice, will be removed in Heist 1.1" #-}
 
 
 ------------------------------------------------------------------------------
@@ -656,32 +676,35 @@ withSplices splice splices runtimeAction =
     splices' = mapV ($runtimeAction) splices
 
 
+------------------------------------------------------------------------------
+{-# INLINE foldMapM #-}
+foldMapM :: (Monad f, Monoid m, Foldable list)
+         => (a -> f m)
+         -> list a
+         -> f m
+foldMapM f =
+  Foldable.foldlM (\xs x -> xs `seq` liftM (xs <>) (f x)) mempty
+
 ------------------------------------------------------------------------------
 -- | Like withSplices, but evaluates the splice repeatedly for each element in
 -- a list generated at runtime.
-manyWithSplices :: Monad n
+manyWithSplices :: (Foldable f, Monad n)
                 => Splice n
                 -> Splices (RuntimeSplice n a -> Splice n)
-                -> RuntimeSplice n [a]
+                -> RuntimeSplice n (f a)
                 -> Splice n
-manyWithSplices splice splices runtimeAction = do
-    p <- newEmptyPromise
-    let splices' = mapV ($ getPromise p) splices
-    chunks <- withLocalSplices splices' mempty splice
-    return $ yieldRuntime $ do
-        items <- runtimeAction
-        res <- forM items $ \item -> putPromise p item >> codeGen chunks
-        return $ mconcat res
+manyWithSplices splice splices runtimeAction =
+    manyWith splice splices mempty runtimeAction
 
 
 ------------------------------------------------------------------------------
 -- | More powerful version of manyWithSplices that lets you also define
 -- attribute splices.
-manyWith :: (Monad n)
+manyWith :: (Foldable f, Monad n)
          => Splice n
          -> Splices (RuntimeSplice n a -> Splice n)
          -> Splices (RuntimeSplice n a -> AttrSplice n)
-         -> RuntimeSplice n [a]
+         -> RuntimeSplice n (f a)
          -> Splice n
 manyWith splice splices attrSplices runtimeAction = do
     p <- newEmptyPromise
@@ -690,27 +713,23 @@ manyWith splice splices attrSplices runtimeAction = do
     chunks <- withLocalSplices splices' attrSplices' splice
     return $ yieldRuntime $ do
         items <- runtimeAction
-        res <- forM items $ \item -> putPromise p item >> codeGen chunks
-        return $ mconcat res
+        foldMapM (\item -> putPromise p item >> codeGen chunks) items
 
 
 ------------------------------------------------------------------------------
 -- | Similar to 'mapSplices' in interpreted mode.  Gets a runtime list of
 -- items and applies a compiled runtime splice function to each element of the
 -- list.
-deferMany :: Monad n
+deferMany :: (Foldable f, Monad n)
           => (RuntimeSplice n a -> Splice n)
-          -> RuntimeSplice n [a]
+          -> RuntimeSplice n (f a)
           -> Splice n
 deferMany f getItems = do
     promise <- newEmptyPromise
     chunks <- f $ getPromise promise
     return $ yieldRuntime $ do
         items <- getItems
-        res <- forM items $ \item -> do
-            putPromise promise item
-            codeGen chunks
-        return $ mconcat res
+        foldMapM (\item -> putPromise promise item >> codeGen chunks) items
 
 
 ------------------------------------------------------------------------------
