@@ -5,6 +5,7 @@
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE MultiParamTypeClasses      #-}
 {-# LANGUAGE ScopedTypeVariables        #-}
+{-# LANGUAGE TypeFamilies               #-}
 {-# LANGUAGE UndecidableInstances       #-}
 
 {-|
@@ -19,32 +20,46 @@ liberating us from the unused writer portion of RWST.
 module Heist.Internal.Types.HeistState where
 
 ------------------------------------------------------------------------------
-import           Blaze.ByteString.Builder
-import           Control.Applicative
-import           Control.Arrow
-import           Control.Monad.CatchIO (MonadCatchIO)
-import qualified Control.Monad.CatchIO as C
-import           Control.Monad.Cont
+import           Blaze.ByteString.Builder      (Builder)
+import           Control.Applicative           (Alternative (..))
+import           Control.Arrow                 (first)
+import           Control.Monad                 (MonadPlus (..), ap)
+import           Control.Monad.Base
+import           Control.Monad.Cont            (MonadCont (..))
 #if MIN_VERSION_mtl(2,2,1)
 import           Control.Monad.Except          (MonadError (..))
 #else
 import           Control.Monad.Error           (MonadError (..))
 #endif
-import           Control.Monad.Reader
-import           Control.Monad.State.Strict
-import           Data.ByteString.Char8 (ByteString)
-import           Data.DList                      (DList)
-import qualified Data.HashMap.Strict as H
-import           Data.HashMap.Strict (HashMap)
-import           Data.HeterogeneousEnvironment   (HeterogeneousEnvironment)
+import           Control.Monad.Fix             (MonadFix (..))
+import           Control.Monad.Reader          (MonadReader (..))
+import           Control.Monad.State.Strict    (MonadState (..), StateT)
+import           Control.Monad.Trans           (MonadIO (..), MonadTrans (..))
+import           Control.Monad.Trans.Control
+import           Data.ByteString.Char8         (ByteString)
+import           Data.DList                    (DList)
+import           Data.HashMap.Strict           (HashMap)
+import qualified Data.HashMap.Strict           as H
+import           Data.HeterogeneousEnvironment (HeterogeneousEnvironment)
 import qualified Data.HeterogeneousEnvironment as HE
 import           Data.Map.Syntax
-import           Data.Monoid
-import           Data.Text (Text)
-import qualified Data.Text as T
-import           Data.Text.Encoding
-import           Data.Typeable
-import qualified Text.XmlHtml as X
+import           Data.Text                     (Text)
+import qualified Data.Text                     as T
+import           Data.Text.Encoding            (decodeUtf8)
+#if MIN_VERSION_base (4,7,0)
+import           Data.Typeable                 (Typeable)
+#else
+import           Data.Typeable                 (TyCon, Typeable(..),
+                                                Typeable1(..), mkTyCon,
+                                                mkTyConApp)
+#endif
+
+#if !MIN_VERSION_base(4,8,0)
+import           Control.Applicative           (Applicative (..), (<$>))
+import           Data.Monoid                   (Monoid(..))
+#endif
+
+import qualified Text.XmlHtml                  as X
 ------------------------------------------------------------------------------
 
 
@@ -150,6 +165,16 @@ isPureChunk _ = False
 type AttrSplice m = Text -> RuntimeSplice m [(Text, Text)]
 
 
+------------------------------------------------------------------------------
+-- | Detailed information about a splice error.
+data SpliceError = SpliceError
+    { spliceHistory      :: [(TPath, Maybe FilePath, Text)]
+    , spliceTemplateFile :: Maybe FilePath
+    , visibleSplices     :: [Text]
+    , spliceMsg          :: Text
+    } deriving ( Show, Eq )
+
+
 ------------------------------------------------------------------------------
 -- | Holds all the state information needed for template processing.  You will
 -- build a @HeistState@ using 'initHeist' and any of Heist's @HeistState ->
@@ -175,6 +200,8 @@ data HeistState m = HeistState {
     , _recurse             :: Bool
     -- | The path to the template currently being processed.
     , _curContext          :: TPath
+    -- | Stack of the splices used.
+    , _splicePath          :: [(TPath, Maybe FilePath, Text)]
     -- | A counter keeping track of the current recursion depth to prevent
     -- infinite loops.
     , _recursionDepth      :: Int
@@ -198,7 +225,7 @@ data HeistState m = HeistState {
     , _splicePrefix        :: Text
 
     -- | List of errors encountered during splice processing.
-    , _spliceErrors        :: [Text]
+    , _spliceErrors        :: [SpliceError]
 
     -- | Whether to throw an error when a tag wih the heist namespace does not
     -- correspond to a bound splice.  When not using a namespace, this flag is
@@ -332,15 +359,44 @@ instance MonadTrans (HeistT n) where
         return (a, s)
 
 
-------------------------------------------------------------------------------
--- | MonadCatchIO instance
-instance MonadCatchIO m => MonadCatchIO (HeistT n m) where
-    catch (HeistT a) h = HeistT $ \r s -> do
-       let handler e = runHeistT (h e) r s
-       C.catch (a r s) handler
-    block (HeistT m) = HeistT $ \r s -> C.block (m r s)
-    unblock (HeistT m) = HeistT $ \r s -> C.unblock (m r s)
+instance MonadBase b m => MonadBase b (HeistT n m) where
+    liftBase = lift . liftBase
 
+#if MIN_VERSION_monad_control(1,0,0)
+instance MonadTransControl (HeistT n) where
+    type StT (HeistT n) a = (a, HeistState n)
+    liftWith f = HeistT $ \n s -> do
+        res <- f $ \(HeistT g) -> g n s
+        return (res, s)
+    restoreT k = HeistT $ \_ _ -> k
+    {-# INLINE liftWith #-}
+    {-# INLINE restoreT #-}
+
+
+instance MonadBaseControl b m => MonadBaseControl b (HeistT n m) where
+     type StM (HeistT n m) a = ComposeSt (HeistT n) m a
+     liftBaseWith = defaultLiftBaseWith
+     restoreM = defaultRestoreM
+     {-# INLINE liftBaseWith #-}
+     {-# INLINE restoreM #-}
+#else
+instance MonadTransControl (HeistT n) where
+    newtype StT (HeistT n) a = StHeistT {unStHeistT :: (a, HeistState n)}
+    liftWith f = HeistT $ \n s -> do
+        res <- f $ \(HeistT g) -> liftM StHeistT $ g n s
+        return (res, s)
+    restoreT k = HeistT $ \_ _ -> liftM unStHeistT k
+    {-# INLINE liftWith #-}
+    {-# INLINE restoreT #-}
+
+
+instance MonadBaseControl b m => MonadBaseControl b (HeistT n m) where
+     newtype StM (HeistT n m) a = StMHeist {unStMHeist :: ComposeSt (HeistT n) m a}
+     liftBaseWith = defaultLiftBaseWith StMHeist
+     restoreM = defaultRestoreM unStMHeist
+     {-# INLINE liftBaseWith #-}
+     {-# INLINE restoreM #-}
+#endif
 
 ------------------------------------------------------------------------------
 -- | MonadFix passthrough instance
@@ -453,7 +509,7 @@ instance (Typeable1 m) => Typeable1 (HeistT n m) where
 -- returns the Node for the @speech@ tag and its children.  @getParamNode >>=
 -- childNodes@ returns a list containing one 'TextNode' containing part of
 -- Hamlet's speech.  @liftM (getAttribute \"author\") getParamNode@ would
--- return @Just "Shakespeare"@.
+-- return @Just \"Shakespeare\"@.
 getParamNode :: Monad m => HeistT n m X.Node
 getParamNode = HeistT $ curry return
 {-# INLINE getParamNode #-}
